@@ -9,11 +9,11 @@ TAPx::Parser::Iterator - Internal TAPx::Parser Iterator
 
 =head1 VERSION
 
-Version 0.50_06
+Version 0.50_07
 
 =cut
 
-$VERSION = '0.50_06';
+$VERSION = '0.50_07';
 
 =head1 SYNOPSIS
 
@@ -41,6 +41,10 @@ Create an iterator.
 
 Iterate through it, of course.
 
+=head2 next_raw()
+
+Iterate raw input without applying any fixes for quirky input syntax.
+
 =head2 is_first()
 
 Returns true if on the first line.  Must be called I<after> C<next()>.
@@ -56,6 +60,7 @@ sub new {
 
     my $ref = ref $thing;
     if ( $ref eq 'GLOB' || $ref eq 'IO::Handle' ) {
+
         # we may eventually allow a 'fast' switch which can read the entire
         # stream into an array.  This seems to speed things up by 10 to 12
         # per cent.  Should not be used with infinite streams.
@@ -81,7 +86,7 @@ package TAPx::Parser::Iterator::FH;
 
 use vars qw($VERSION @ISA);
 @ISA     = 'TAPx::Parser::Iterator';
-$VERSION = '0.50_06';
+$VERSION = '0.50_07';
 
 sub new {
     my ( $class, $thing ) = @_;
@@ -118,7 +123,7 @@ sub exit     { $_[0]->{exit} }
 sub is_first { $_[0]->{is_first} }
 sub is_last  { $_[0]->{is_last} }
 
-sub next {
+sub next_raw {
     my $self = shift;
     my $fh   = $self->{fh};
 
@@ -144,27 +149,33 @@ sub next {
         }
     }
 
+    return $line;
+}
+
+sub next {
+    my $self = shift;
+    my $line = $self->next_raw;
+
     # vms nit:  When encountering 'not ok', vms often has the 'not' on a line
     # by itself:
     #   not
     #   ok 1 - 'I hate VMS'
     if ( defined $line && $line =~ /^\s*not\s*$/ ) {
-        $line .= ($self->next || '');
+        $line .= ( $self->next_raw || '' );
     }
-    
     return $line;
 }
 
 sub _finish {
     my $self = shift;
-    
+
     my $status = $?;
-    
+
     # If we have a subprocess we need to wait for it to terminate
     if ( defined $self->{pid} ) {
         if ( $self->{pid} == waitpid( $self->{pid}, 0 ) ) {
             $status = $?;
-        };
+        }
     }
 
     close $self->{fh};
@@ -181,7 +192,7 @@ package TAPx::Parser::Iterator::ARRAY;
 
 use vars qw($VERSION @ISA);
 @ISA     = 'TAPx::Parser::Iterator';
-$VERSION = '0.50_06';
+$VERSION = '0.50_07';
 
 sub new {
     my ( $class, $thing ) = @_;
@@ -203,4 +214,6 @@ sub next {
     return $self->{array}->[ $self->{idx}++ ];
 }
 
+sub next_raw { shift->next }
+
 1;
